#include "common.h"
#include "serial.h"

#include <avr/io.h>
#include <avr/interrupt.h>

#define MEGA88

#ifndef F_CPU
/* In neueren Version der WinAVR/Mfile Makefile-Vorlage kann
   F_CPU im Makefile definiert werden, eine nochmalige Definition
   hier wuerde zu einer Compilerwarnung fuehren. Daher "Schutz" durch
   #ifndef/#endif 
 
   Dieser "Schutz" kann zu Debugsessions fhren, wenn AVRStudio 
   verwendet wird und dort eine andere, nicht zur Hardware passende 
   Taktrate eingestellt ist: Dann wird die folgende Definition 
   nicht verwendet, sondern stattdessen der Defaultwert (8 MHz?) 
   von AVRStudio - daher Ausgabe einer Warnung falls F_CPU
   noch nicht definiert: */

#error "F_CPU war noch nicht definiert, bitte nachholen"
// z.B.: #define F_CPU 8000000UL    // Systemtakt in Hz - Definition als unsigned long beachten >> Ohne ergeben Fehler in der Berechnung
#endif


#if DEBUG > 0
  #define vartype static
#endif

  typedef struct 
  {
    uint16_t ubrr_val;
    uint16_t baud_real;
    uint16_t baud_error;
  } baudcalc_t;

  //enum baudcalc_t
  //{
  #define BAUDCALC_NORMAL   1
	#define BAUDCALC_PLUS1    2
	#define BAUDCALC_X2       3
	#define BAUDCALC_X2_PLUS1 4
  //};

uint16_t USART_u16Ubrr_val(uint16_t baudrate);
void USART_baudcalc(uint16_t baudrate, baudcalc_t* pBaudcalcValues, uint8_t type);


///////////////////////////////////////////////////////////////////////////////
// @function :
void USART_baudcalc(uint16_t baudrate, baudcalc_t* pBaudcalcValues, uint8_t type)

//
// abstract : calculate the value of the UBRR register from the desired
//            baud rate  
//
// @param   : 
//
// @return  :     
//
// @history :
//
// name   date        action
// ----------------------------------------------------------------------------
// ode    14.07.2009  implementation start
///////////////////////////////////////////////////////////////////////////////
{
  vartype uint32_t numerator;
  vartype uint32_t denominator;
  vartype uint16_t my_baudrate;
  vartype baudcalc_t* pMyBaudcalcValues;
  vartype uint8_t speed;
  vartype uint8_t offset; 

  my_baudrate = baudrate;
  pMyBaudcalcValues = pBaudcalcValues;

  switch (type)
  {
    case BAUDCALC_NORMAL:
	  {
	    speed  = 1;
	    offset = 1; // always >= 1 since UBRR starts counting at 0
	  }
	  break;

	  case BAUDCALC_PLUS1:
    {
	    speed  = 1;
      offset = 2; // try if error is lower with another + 1
	  }
	  break;

	  case BAUDCALC_X2:
    {
	    speed  = 2;
	    offset = 2; // always >=2 in serial x2 mode
	  }
	  break;

	  case BAUDCALC_X2_PLUS1:
    {
	    speed  = 2;
      offset = 3; // try if error is lower with another + 1
	  }
	  break;

	  default:
	    ;
  }


  // Berechnungen
  numerator   = F_CPU + (uint32_t) my_baudrate*8;
  denominator = (uint32_t) my_baudrate*16/speed;
  pBaudcalcValues->ubrr_val    = numerator/denominator -1; // clever runden
  pBaudcalcValues->baud_real   = (F_CPU/(16/speed*( (uint32_t) pBaudcalcValues->ubrr_val+offset)));              // Reale Baudrate
  pBaudcalcValues->baud_error  = (( (uint32_t) pBaudcalcValues->baud_real * 1000)/my_baudrate);          // Fehler in Promille, 1000 = kein Fehler.
}
//
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// @function :
uint16_t USART_u16Ubrr_val(uint16_t baudrate)

//
// abstract : find the UBRR register value and x2 mode bit value from the 
//            desired baud rate that gives the closest match 
//
// @param   : 
//
// @return  : type/variable              : abstract    
//             0                          : always     
//
// @history :
//
// name   date        action
// ----------------------------------------------------------------------------
// ode    13.07.2009  implementation start
///////////////////////////////////////////////////////////////////////////////
{

  baudcalc_t baudcalcValues;

  USART_baudcalc(baudrate, &baudcalcValues, BAUDCALC_NORMAL);
  /*
  USART_baudcalc(baudrate, &baudcalcValues, BAUDCALC_PLUS1);
  USART_baudcalc(baudrate, &baudcalcValues, BAUDCALC_X2);
  USART_baudcalc(baudrate, &baudcalcValues, BAUDCALC_X2_PLUS1);
  */

#if DEBUG >0
  vartype uint16_t my_baudrate;
  my_baudrate = baudrate;
#endif

  if (  (baudcalcValues.baud_error < BAUD_ERROR_LIMIT_LO) 
      ||(baudcalcValues.baud_error > BAUD_ERROR_LIMIT_HI) )
    return 0xFFFF; // error
  else
    return baudcalcValues.ubrr_val;
}
//
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// @function :

void USART_vInit(uint8_t usartnr, uint16_t baudrate)
//
// abstract:  UART0 initialize
//            desired baud rate: 115200
//            actual: baud rate:125000 (7,8%)
//            char size: 8 bit
//            parity: Disabled
//
// @param   : access/type/variable   : abstract
//            void                   : no parameters
//
// @return   : type/variable              : abstract    
//             0                          : always     
//
// @history  :
//
// name   date        action
// ----------------------------------------------------------------------------
// ode    08.07.2009  implementation start
///////////////////////////////////////////////////////////////////////////////
{
  if (usartnr == 0)
  {
  #ifdef MEGA88
  uint16_t ubrr_val = USART_u16Ubrr_val(baudrate);

  // Set baud rate
  UBRR0H = (uint8_t) (ubrr_val>>8);
  UBRR0L = (uint8_t) ubrr_val & 0xFF;

  // Set frame format to 8 data bits, no parity, 1 stop bit
  UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);

#ifdef UBRR_VALx2
  // Set baud rate
  UBRR0H = (uint8_t) (UBRR_VALx2>>8);
  UBRR0L = (uint8_t) UBRR_VALx2 & 0xFF;

  UCSR0A |= (1<<U2X0);
#endif
  // Enable receiver and transmitter
  UCSR0B = (1<<TXEN0);
  
  #else // MEGA32
  UCSRB = 0x00;
  UCSRA = 0x00;
  UCSRC = (URSEL << 1)| 0x06;
  UBRRL = 0x07;
  UBRRH = 0x00;
  UCSRB = 0x18;  
  #endif



  }
  else
  {
  }


}
// 
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// @function :

void USART_SendByte(char_t data)

//
// abstract:  UART0 SendByte
//
//            USART transmitter is enabled by setting TXEN bit in UCSRB register. 
//            Normal port operations is overridden by USART (XCK pin in synchronous 
//            operations will be overridden too).
//
// @param   : access/type/variable   : abstract
//            void                   : no parameters
//
// @return   : type/variable              : abstract    
//             0                          : always     
//
// @history  :
//
// name   date        action
// ----------------------------------------------------------------------------
// ode    08.07.2009  implementation start
///////////////////////////////////////////////////////////////////////////////
{

  // Wait if a byte is being transmitted
  #ifdef MEGA88

  while((UCSR0A&(1<<UDRE0)) == 0)
    ;
  // Transmit data
  UDR0 = data;

  #else // MEGA32

  /* Wait for empty transmit buffer */
  while ( !( UCSRA & (1<<UDRE)) )
    ;
  /* Put data into buffer, sends the data */
  UDR = data;

  #endif

}
// 
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// @function :

char_t USART_vReceiveByte(void)
//
// abstract:  UART0 SendByte
//
//            USART transmitter is enabled by setting TXEN bit in UCSRB register. 
//            Normal port operations is overridden by USART (XCK pin in synchronous 
//            operations will be overridden too).
//
// @param   : access/type/variable   : abstract
//            void                   : no parameters
//
// @return   : type/variable              : abstract    
//             0                          : always     
//
// @history  :
//
// name   date        action
// ----------------------------------------------------------------------------
// ode    08.07.2009  implementation start
///////////////////////////////////////////////////////////////////////////////
{
  #ifdef MEGA88
  // The USART Receiver is enabled by writing the Receive Enable (RXEN) bit in the UCSRB Register to one. When the Receiver is enabled, the normal pin operation of the RxD pin is overridden by the USART and given the function as the Receivers serial input. The baud rate, mode of operation and frame format must be set up before any transfer operations.

  // Wait until a byte has been received

  while((UCSR0A&(1<<RXC0)) == 0) 
    ;

  // Return received data
  return UDR0;
  
  #else // MEGA32

  #endif

}
// 
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// @function :

void USART_SendString(char_t* data)
//
// abstract:  UART0 SendSend String with maximum lenth of 64 Bytes
//
// @param   : 
//
// @return   :    
//
// @history  :
//
// name   date        action
// ----------------------------------------------------------------------------
// ode    14.07.2009  implementation start
///////////////////////////////////////////////////////////////////////////////
{

  // Wait if a byte is being transmitted
  uint8_t strLen = 0;

  while(   (*data != '\0') 
        && (strLen <= 64))
  {
    USART_SendByte(*data);
	  data++;
	  strLen++;
  } 

}
// 
///////////////////////////////////////////////////////////////////////////////

// eof 
